from requests.auth import AuthBase
import logging
import json
import os
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)


class BearerAuth(AuthBase):
    """Authenticate with Blackduck hub using access token"""

    def __init__(self, session):
        """
        Args:
            session (requests.session): requests session to authenticate
        """

        self.blackduck_url = os.environ.get("BCOM_BDUCK_URL")
        self.api_token = os.environ.get("BCOM_BDUCK_API_TOKEN")
        self.session = session
        self.bearer_token = None
        self.csrf_token = None
        self.valid_until = datetime.now()

    def __call__(self, request):
        if not self.bearer_token or datetime.now() > self.valid_until - timedelta(
            minutes=5
        ):
            self.authenticate()

        request.headers.update(
            {
                "authorization": f"bearer {self.bearer_token}",
                "X-CSRF-TOKEN": self.csrf_token,
            }
        )

        return request

    def authenticate(self):
        print(f"** INFO: Authenticating to BlackDuck API ..")
        response = self.session.post(
            url=f"{self.blackduck_url}/api/tokens/authenticate",
            headers={"Authorization": f"token {self.api_token}"},
        )

        if response.status_code == 200:
            try:
                content = response.json()
                self.bearer_token = content["bearerToken"]
                self.csrf_token = response.headers["X-CSRF-TOKEN"]
                self.valid_until = datetime.now() + timedelta(
                    milliseconds=int(content["expiresInMilliseconds"])
                )

                logger.info(
                    f"success: auth granted until {self.valid_until.astimezone()}"
                )

                return
            except (json.JSONDecodeError, KeyError):
                logger.exception(
                    "HTTP response status code 200 but unable to obtain bearer token"
                )

        if response.status_code == 401:
            logger.error("HTTP response status code = 401 (Unauthorized)")

            try:
                logger.error(response.json()["errorMessage"])
            except (json.JSONDecodeError, KeyError):
                logger.exception("unable to extract error message")
                logger.error("HTTP response headers: %s", response.headers)
                logger.error("HTTP response text: %s", response.text)
            raise RuntimeError("Unauthorized access token", response)

        logger.error("Unhandled HTTP response")
        logger.error("HTTP response status code %i", response.status_code)
        logger.error("HTTP response headers: %s", response.headers)
        logger.error("HTTP response text: %s", response.text)

        raise RuntimeError("Unhandled HTTP response", response)
