import os
import requests
from authentication import BearerAuth


class BlackDuckCore:
    def __init__(self):
        """
        Args:
            base_url (string): BlackDuck base URL
            session (requests.session): requests session to authenticate
            token (string): BlackDuck API token
        """

        self.blackduck_url = os.environ.get("BCOM_BDUCK_URL")
        self.api_token = os.environ.get("BCOM_BDUCK_API_TOKEN")
        self.session = requests.Session()
        self.auth = BearerAuth(self.session)

    def get_project_id(self, project_name):
        url = f"{self.blackduck_url}/api/projects?q=name:{project_name}"

        print(f"** INFO: Getting project ID for {project_name} ..")
        response = self.session.get(url, auth=self.auth)

        response.raise_for_status()

        projects = response.json().get("items", [])

        if projects:
            _project_id = projects[0]["_meta"]["href"].split("/")[-1]
            print(f"** INFO: Project ID: {_project_id} ..")

            return _project_id
        else:
            raise ValueError("** ERROR: Project not found")

    def get_projects(self, project_name, limit=1, parameters={}):
        if limit:
            parameters.update({"limit": limit})

        url = f"{self.blackduck_url}/{parameters}"
        headers = {}
        headers["Accept"] = "application/vnd.blackducksoftware.project-detail-4+json"

        print(f"** INFO: Retrieving projects using url {url}")

        response = self.session.get(url, headers=headers, verify=True, auth=self.auth)
        projects = response.json()

        return projects

    def get_project_by_name(self, project_name):
        project_list = self.get_projects(parameters={f"'q': 'name:{project_name}'"})

        for project in project_list["items"]:
            if project["name"] == project_name:
                return project

    def project_exists(self, project_name):
        """
        Check if the project exists in BlackDuck.
        """

        url = f"{self.blackduck_url}/api/projects"

        print(f"** INFO: Checking if project {project_name} exists in BlackDuck ..")
        response = self.session.get(f"{url}?q=name:{project_name}", auth=self.auth)
        response.raise_for_status()

        return response.json().get("totalCount", 0) > 0

    def get_project_versions(self, project_id):
        url = f"{self.blackduck_url}/api/projects/{project_id}/versions"

        print(f"** INFO: Getting project version for {url} ..")
        response = self.session.get(url, auth=self.auth)

        response.raise_for_status()
        return response.json().get("items", [])

    def get_project_distribution(self, version_url):
        print(f"** INFO: Getting project distribution for {version_url} ..")
        response = self.session.get(version_url, auth=self.auth)

        response.raise_for_status()

        _project_distribution = response.json().get("distribution", "EXTERNAL")
        print(f"** INFO: Project distribution: {_project_distribution} ..")

        return _project_distribution

    def update_version_phase(self, version_url, project_version_name, phase):
        distribution = self.get_project_distribution(version_url)

        data = {
            "phase": phase,
            "versionName": project_version_name,
            "distribution": distribution,
        }

        print(
            f"** INFO: Updating project version phase for {version_url} to {phase} .."
        )
        response = self.session.put(version_url, json=data, auth=self.auth)

        response.raise_for_status()
        return response.json()
