#! /usr/bin/env python3

"""PSEC BlackDuck Scanner.

This script uses BlackDuck to perform licensing,
vulnerabilities, and binary analysis against files and docker images.
"""

import argparse
import os
import re
import sys
import subprocess
import shutil
import requests
import yaml
from requests.auth import HTTPBasicAuth
from urllib.parse import urlparse
from core import BlackDuckCore
from aws import AWS

REP_PATH_CLI = os.environ.get("REP_PATH_CLI", ".")
BCOM_BDUCK_URL = os.environ.get("BCOM_BDUCK_URL")
BCOM_BDUCK_API_TOKEN = os.environ.get("BCOM_BDUCK_API_TOKEN")


def validate_variables():
    """Validates the presence of required environment variables.

    Raises:
        ValueError: if any required environment variable is not set.
    """

    required_vars = ["BCOM_BDUCK_URL", "BCOM_BDUCK_API_TOKEN"]

    print(f"** INFO: Validating required variables ..")

    for var in required_vars:
        if not os.environ.get(var):
            raise ValueError(f"** ERROR: {var} environment variable is not set.")


def download_file(url: str):
    """Downloads target file to be scanned.

    Args:
        url (str): URL of the file to be downloaded.

    Raises:
        RequestException: if the request fails.

    Returns:
        file_path (str): path to the file to be scanned.
    """

    session = requests.Session()
    auth = None

    try:
        if "usw1.packages.broadcom.com" in url:
            auth = HTTPBasicAuth(os.getenv("ARTIFACTORY_USER"), os.getenv("ARTIFACTORY_PASSWORD"))

        file_path = os.path.join(os.getcwd(), os.path.basename(urlparse(url).path))

        with session.get(url, auth=auth, stream=True) as s_response:
            with open(file_path, "wb") as file:
                for chunk in s_response.iter_content(chunk_size=8192):
                    file.write(chunk)

        print(f"** INFO: File downloaded and saved at: {file_path}")
        return file_path
    except requests.exceptions.RequestException as e:
        print("** ERROR: Failed to download the target file!")
        sys.exit(1)


def read_yaml(project_groups, project_name: str):
    """Traverses the dictionary and returns the project name or None.

    Args:
        project_groups (dict): dict containing all project groups.
        project_name (str): the name of the project to search for.

    Returns:
        project_group (str): the project group name.
    """

    if isinstance(project_groups, dict):
        for key, value in project_groups.items():
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        result = read_yaml(item, project_name)

                        if result:
                            return result
                    elif project_name.lower() in item.lower():
                        return key
            elif isinstance(value, dict):
                result = read_yaml(value, project_name)

                if result:
                    return result
    return None


def get_project_group(project_name: str):
    """Determines the project group name.

    Searches for project_name in the mapping file.
    If no matches are found, project_name is set to a default value.

    Args:
        project_name (str): the name of the project to search for.

    Raises:
        Exception: if something fails.

    Returns:
        project_group (str): the project group name.
    """

    mapping_file = os.path.join(os.getcwd(), "blackduck/projects_mapping.yml")

    try:
        with open(mapping_file, 'r') as f:
            project_groups = yaml.safe_load(f)

        project_group = read_yaml(project_groups, project_name)

        if project_group is None:
            project_group = "CB Cloud (CBC)"
    except Exception as exc:
        print(f"** ERROR: An unexpected error occurred: {exc}")
        sys.exit(1)

    return project_group


def run_synopsys_detect(
    project_name: str,
    project_version: str,
    project_group: str,
    DETECT_TOOLS: str,
    EXCLUDED_TOOLS: str,
    LOGGING_LEVEL: str,
    target: str,
):
    """Runs BlackDuck Detect via shell command.

    Invokes BlackDuck Detect and scans a given target
    using several different detectors.

    Args:
        project_name (str): name of the project in BlackDuck HUB.
        project_version (str): version of the project in BlackDuck HUB.
        project_group (str): name of the project group in BlackDuck HUB.
        DETECT_TOOLS (str): Detect tools to include in the scan.
        EXCLUDED_TOOLS (str): Detect tools to exclude from the scan.
        target (str): file, directory or container image to be scanned.

    Raises:
        CalledProcessError: If the subprocess command fails.

    Returns:
        file_path (str): Path to the file to be scanned.
    """

    BDUCK_LOGGING_LEVEL = LOGGING_LEVEL or "INFO"
    BDUCK_DETECT_TOOLS = DETECT_TOOLS or "BINARY_SCAN"
    BDUCK_EXCLUDED_TOOLS = (
        EXCLUDED_TOOLS
        or "DETECTOR,SIGNATURE_SCAN,IMPACT_ANALYSIS,DOCKER,BAZEL,IAC_SCAN,CONTAINER_SCAN,THREAT_INTEL"
    )
    project_name = project_name.lower()
    project_version = project_version.lower()
    project_group = project_group or get_project_group(project_name)
    target = target
    image_name_pattern = re.compile(
        r"^([a-z0-9.-]+(:[a-zA-Z0-9.-]+)?(/[a-zA-Z0-9.-]+)+(:[a-zA-Z0-9.-]+)?)$"
    )
    blackduck_core = BlackDuckCore()

    if not blackduck_core.project_exists(project_name):
        print("** ERROR: The BlackDuck project does not exist. Exiting ..")
        sys.exit(1)

    cmd = [
        "java",
        "-jar",
        "/root/synopsys-detect.jar",
        f"--blackduck.url={BCOM_BDUCK_URL}",
        f"--blackduck.api.token={BCOM_BDUCK_API_TOKEN}",
        "--blackduck.trust.cert=true",
        f"--detect.project.group.name={project_group}",
        f"--detect.project.name={project_name}",
        f"--detect.project.version.name={project_version}",
        f"--detect.tools={BDUCK_DETECT_TOOLS}",
        f"--detect.tools.excluded={BDUCK_EXCLUDED_TOOLS}",
        f"--logging.level.com.synopsys.integration={BDUCK_LOGGING_LEVEL}",
        "--detect.risk.report.pdf=false",
        "--detect.wait.for.results=false",
        "--detect.policy.check.fail.on.severities=BLOCKER,CRITICAL",
        "--detect.cleanup=true",
        "--detect.clone.project.version.latest=true",
        "--detect.project.version.phase=RELEASED",
    ]

    if os.path.isdir(target):
        print(f"** ERROR: Provide the full path to the binary file. A path to a directory was provided instead.")
        sys.exit(1)
    elif os.path.isfile(target):
        destination_path = os.path.join(REP_PATH_CLI, os.path.basename(target))

        if not os.path.exists(destination_path):
            try:
                target = shutil.copy2(target, f"{REP_PATH_CLI}/")
                cmd.append(f"--detect.binary.scan.file.path={target}")
            except Exception as e:
                print(f"** ERROR: Failed to copy the target file: {e}")
                sys.exit(1)
        else:
            print("** INFO: Target file already exists in the destination. Proceeding ..")
            cmd.append(f"--detect.binary.scan.file.path={destination_path}")
    elif target.startswith(("http://", "https://")):
        target = download_file(target)

        try:
            target = shutil.copy2(target, f"{REP_PATH_CLI}/")
            cmd.append(f"--detect.binary.scan.file.path={target}")
        except Exception as e:
            print(f"** ERROR: Failed to copy the target file: {e}")
            sys.exit(1)
    elif image_name_pattern.match(target):
        if not project_name:
            pn_index = cmd.index(f"--detect.project.name={project_name}")
            pn = target.rsplit(":", 1)[0]
            cmd[pn_index] = f"--detect.project.name={pn}"
        if not project_version:
            pv_index = cmd.index(f"--detect.project.version.name={project_version}")
            pv = target.split(":")[-1]
            cmd[pv_index] = f"--detect.project.version.name={pv}"

        cmd.append(f"--detect.docker.image={target}")
        cmd.append(
            f"--detect.docker.passthrough.imageinspector.service.distro.default=centos"
        )

        if ".dkr.ecr." in target.lower():
            AWS.login_to_ecr(target)
    else:
        print("** ERROR: Provide a file URL, local file path, or container image name.")
        sys.exit(1)

    try:
        subprocess.run(cmd)
    except subprocess.CalledProcessError as e:
        print(f"** ERROR: Command execution failed: {e}")
        sys.exit(1)

    project_id = blackduck_core.get_project_id(project_name)
    versions = blackduck_core.get_project_versions(project_id)

    if len(versions) < 1:
        print("** ERROR: Not enough versions to find a previous version")
        sys.exit(1)

    released_versions = [v for v in versions if v["phase"].lower() == "released"]

    if released_versions:
        released_versions.sort(key=lambda v: v["createdAt"])

        latest_released_version = released_versions[-1]
        latest_released_version_url = latest_released_version["_meta"]["href"]

        blackduck_core.update_version_phase(
            latest_released_version_url, project_version, "DEPRECATED"
        )
    else:
        print("** INFO: No versions in RELEASED phase found! The previous version will NOT be marked as DEPRECATED.")


def main():
    validate_variables()

    parser = argparse.ArgumentParser(description="PSEC BlackDuck Scanner")

    parser.add_argument(
        "-pn",
        "--project_name",
        help="Project name in BlackDuck",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-pv",
        "--project_version",
        help="Project version in BlackDuck",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-pg",
        "--project_group",
        help="Project parent group name in BlackDuck",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-dt",
        "--detect_tools",
        help="Detect tools to include in the scan",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-et",
        "--excluded_tools",
        help="Detect tools to exclude from the scan",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-t",
        "--target",
        help="File URL, local file path, or full container image URL",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-ll",
        "--logging_level",
        help="BlackDuck logging level",
        required=False,
        type=str,
    )

    args = parser.parse_args()
    project_name = args.project_name
    project_version = args.project_version
    project_group = args.project_group
    DETECT_TOOLS = args.detect_tools
    EXCLUDED_TOOLS = args.excluded_tools
    LOGGING_LEVEL = args.logging_level
    target = args.target

    print(f"** INFO: Initiating BlackDuck scan activities ..")

    for t in target.split():
        run_synopsys_detect(
            project_name,
            project_version,
            project_group,
            DETECT_TOOLS,
            EXCLUDED_TOOLS,
            LOGGING_LEVEL,
            t,
        )

    print("** INFO: BlackDuck scan activities are now complete.")


if __name__ == "__main__":
    main()
